import pytest

from django_rubble.fields.db_fields import SimplePercentageField
from django_rubble.forms.db_forms import (
    SimplePercentageField as SimplePercentageFormField,
)


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (None, None),  # decimal_places is None
        (2, 0),  # Percent has no decimal places
    ],
)
def test_simple_percentage_field(value: int | None, expected: int | None):
    field = SimplePercentageField(max_digits=3, decimal_places=value)

    assert field.humanize_decimal_places == expected
    assert isinstance(field.formfield(), SimplePercentageFormField)  # pyright: ignore[reportUnknownMemberType]



import logging
import sys
from enum import Enum
from typing import TYPE_CHECKING, Any, Literal

from django import get_version
from django.apps import apps
from django.conf import settings
from django.views import View
from neapolitan.views import Role
from pydantic import BaseModel, field_validator, model_serializer

from django_rubble.context import PydanticBase

if TYPE_CHECKING:
    from django_stubs_ext import StrOrPromise

logger = logging.getLogger(__name__)


class DuplicateAppError(Exception):
    pass


class NotRegisteredError(Exception):
    pass


class ModelLink(BaseModel):
    name: str
    role: Role
    view_class: type[View]
    url: str

    def get_url(self):
        return self.role.reverse(self.view_class)

    @model_serializer
    def get_dict(self):
        return {
            "name": self.name,
            "role": self.role,
            "view_class": self.view_class,
            "url": self.get_url(),
        }


class ModelViewsConfiguration(BaseModel):
    verbose_name: str
    views_class: type[View]
    menu_roles: list[Role] = [Role.LIST, Role.CREATE]


class ProjectApp(BaseModel):
    name: str | Any
    registered_models: list[tuple[str, type[View]] | ModelViewsConfiguration] = []

    @field_validator("name", mode="before")
    @classmethod
    def str_from_str_or_promise(cls, value: "StrOrPromise | str") -> str:
        return str(value) if not isinstance(value, str) else value

    def get_model(self, model_name: str):
        return apps.get_model(self.name, model_name)

    def get_actions(self) -> list[ModelLink]:
        action_links = []
        for model in self.registered_models:
            views_config = (
                model
                if isinstance(model, ModelViewsConfiguration)
                else ModelViewsConfiguration(
                    verbose_name=model[0], views_class=model[1]
                )
            )
            if Role.LIST in views_config.menu_roles:
                list_link = ModelLink(
                    name=views_config.verbose_name,
                    role=Role.LIST,
                    view_class=views_config.views_class,
                    url="",
                )
                action_links.append(list_link)
            if Role.CREATE in views_config.menu_roles:
                create_link = ModelLink(
                    name=views_config.verbose_name,
                    role=Role.CREATE,
                    view_class=views_config.views_class,
                    url="",
                )
                action_links.append(create_link)

        return action_links

    def register(self, model: tuple[str, type[View]] | ModelViewsConfiguration):
        """Register a model to the app.

        Arguments:
          model: tuple of `(model_name, view_class)`
        """
        if not isinstance(model, ModelViewsConfiguration):
            msg = f"{model[0]} being cast as ModelViewsConfiguration"
            logger.debug(msg)

            model = ModelViewsConfiguration(verbose_name=model[0], views_class=model[1])
        self.registered_models.append(model)


class RegistryBase(type):
    REGISTRY = {}

    def __new__(cls, name: str, bases: tuple[type, ...], attrs: dict[str, Any]):
        new_cls = type.__new__(cls, name, bases, attrs)
        cls.REGISTRY[new_cls.__name__.lower()] = new_cls
        return new_cls

    @classmethod
    def get_registry(cls):
        return dict(cls.REGISTRY)


def base_url_format() -> str:
    return getattr(settings, "RUBBLE_BASE_URL_FORMAT", "{app_name}/{model_name}/")


class RoleConfig(PydanticBase):
    path_fragment: str | None = None
    name_suffix: str

    def render_fragment(self, **kwargs):
        return self.path_fragment.format(**kwargs) or ""

    def render_name(self, model_name: str):
        return f"{model_name}_{self.name_suffix}"


class RoleMapping(Enum):
    LIST = RoleConfig(name_suffix="list")
    CREATE = RoleConfig(path_fragment="new/", name_suffix="create")
    READ = RoleConfig(path_fragment="{lookup_kwarg}/", name_suffix="detail")
    UPDATE = RoleConfig(path_fragment="{lookup_kwarg}/edit/", name_suffix="update")
    DELETE = RoleConfig(path_fragment="{lookup_kwarg}/delete/", name_suffix="delete")

    def __get__(self, *_):
        return self.value


def url_builder(
    app_name: str, model_name: str, view_mapping: RoleConfig = RoleMapping.LIST
):
    return (
        f"{app_name}/{model_name}/{view_mapping.render_fragment(model_name=model_name)}"
    )


class ViewRegistryBase(metaclass=RegistryBase):
    pass


class ProjectRegistry(BaseModel):
    apps: list[ProjectApp] = []
    _django_version: str | None = None
    _python_version: (
        tuple[int, int, int, Literal["alpha", "beta", "candidate", "final"], int] | None
    ) = None
    project_version: str = ""

    @property
    def django_version(self):
        if self._django_version is None:
            self._django_version = get_version()

        return self._django_version

    @property
    def python_version(self) -> str:
        if self._python_version is None:
            self._python_version = sys.version_info
        # "major.minor.micro_releaselevel" format e.g. "3.9.7_final"
        return (
            f"{self._python_version.major}."
            f"{self._python_version.minor}.{self._python_version.micro}_"
            f"{self._python_version.releaselevel}"
        )

    def register(self, app: ProjectApp | tuple[str, tuple[str, type[View]]]):
        """Registers a model/app with the project.

        Arguments:
          app: definition of project "application", either as ProjectApp object or as
          a tuple of `(app_name, (model_name, view_class))`
        """
        if not isinstance(app, ProjectApp):
            app_name, app_model = app
            if not self.is_registered(app_name):
                app = ProjectApp(name=app_name)
            else:
                app = self.get_app_by_name(app_name)
            app.register(app_model)
        self.apps.append(app)

    def is_registered(self, app_name: str) -> bool:
        registered_apps = self.get_app_names()
        return app_name in registered_apps

    def get_app_names(self):
        return [app.name for app in self.apps]

    def get_app_by_name(self, app_name: str) -> ProjectApp:
        named_apps = [app for app in self.apps if app_name in app.name]
        if len(named_apps) > 1:
            raise DuplicateAppError
        if len(named_apps) != 1:
            raise NotRegisteredError
        return named_apps[0]

    def get_context_data(self, *args, **kwargs):
        return [
            {
                "app_name": app.name,
                "views": [action.model_dump() for action in app.get_actions()],
            }
            for app in self.apps
        ]



from typing import Literal

try:
    from infisical_client import (
        AuthenticationOptions,
        ClientSettings,
        GetSecretOptions,
        InfisicalClient,
        UniversalAuthMethod,
    )
except ImportError as e:
    msg = 'Unable to import infisical_client, did you install the "secrets" extra?'
    raise ImportError(msg) from e
from os import getenv

from pydantic import BaseModel, SecretStr


class LongSecretStr(SecretStr):
    def __repr__(self):
        value = self.get_secret_value()
        return f"LongSecretStr('{value[:5]}***{value[-5:]}')"


class Secrets(BaseModel):
    """Safely retrieve secrets from Infisical.

    Does not expose the client secret in the repr.

    Attributes:
        client_id (str): The Infisical client ID. Defaults to the INFISICAL_CLIENT_ID
        client_secret (LongSecretStr): The Infisical client secret. Defaults to the
            INFISICAL_CLIENT_SECRET
        project_id (str): The Infisical project ID. Defaults to the INFISICAL_PROJECT_ID
        environment (Literal["prod", "dev", "staging"]): The Infisical environment.
            Defaults to the INFISICAL_ENVIRONMENT or "dev"
        debug (bool): Whether to run in debug mode. Defaults to False
    """

    client_id: str = getenv("INFISICAL_CLIENT_ID")
    client_secret: LongSecretStr = LongSecretStr(getenv("INFISICAL_CLIENT_SECRET"))
    project_id: str = getenv("INFISICAL_PROJECT_ID")
    environment: Literal["prod", "dev", "staging"] = getenv(
        "INFISICAL_ENVIRONMENT", default="dev"
    )
    debug: bool = False
    _client: InfisicalClient | None = None

    def create_client(self):
        client_secret = self.client_secret.get_secret_value()
        universal_auth = UniversalAuthMethod(
            client_id=self.client_id, client_secret=client_secret
        )

        self._client = InfisicalClient(
            ClientSettings(auth=AuthenticationOptions(universal_auth=universal_auth)),
            debug=self.debug,
        )

    @property
    def client(self):
        if self._client is None:
            self.create_client()
        return self._client

    def get_secret(
        self, secret_name: str, *, environment: str | None = None, mask: bool = False
    ):
        """Retrieve a secret from Infisical.

        Args:
            secret_name: The name of the secret to retrieve.
            environment: The environment to retrieve the secret from. Defaults to the
                environment set in the class.
            mask: Whether to mask the secret in the repr. Defaults to False.
        """
        if mask:
            msg = "Masking is not yet implemented"
            raise NotImplementedError(msg)
        if self._client is None:
            self.create_client()
        if environment is None:
            environment = self.environment
        return self._client.getSecret(
            options=GetSecretOptions(
                environment=environment,
                project_id=self.project_id,
                secret_name=secret_name,
            )
        )



from decimal import Decimal
from typing import TypeVar

import pytest

from django_rubble.utils.numeric_utils import (
    any_to_float,
    is_number,
    ratio_to_whole,
    trim_trailing_zeros,
    whole_to_ratio,
)


@pytest.mark.parametrize(
    ("number", "expected"),
    [
        (10, True),
        (3.14, True),
        ("42", True),
        ("3.14", True),
        ("hello", False),
        (None, False),
    ],
)
def test_is_number(number: int | float | str | None, expected: bool):
    assert is_number(number) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (10, Decimal("10")),
        (3.14, Decimal("3.14")),
        ("42", Decimal("42")),
        ("3.1400", Decimal("3.14")),
    ],
)
def test_trim_trailing_zeros(value: int | float | str, expected: Decimal):
    assert trim_trailing_zeros(value) == expected
    assert Decimal(str(value)).normalize() == expected


@pytest.mark.parametrize(
    ("value", "default", "expected"),
    [
        ("test", 0, 0),
        (5, 0, 5),
        ("5.5", 2.3, 5.5),
    ],
)
def test_value_as_float(value: str | float, default: float, expected: float):
    assert any_to_float(value, default) == expected


def test_any_to_float_exception():
    with pytest.raises(TypeError) as excinfo:
        _ = any_to_float(4.2, "test")  # pyright: ignore[reportArgumentType]
    assert str(excinfo.value) == "Default must be of type `float` [test]"


_T = TypeVar("_T", int, float, Decimal)


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        (0.03, 3),
        (Decimal("1"), Decimal("100")),
        (1, 100),
    ],
)
def test_ratio_to_whole(input_value: float | Decimal, expected: float | Decimal):
    assert ratio_to_whole(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        (3, 0.03),
        (100, 1),
        (Decimal("99"), Decimal("0.99")),
    ],
)
def test_whole_to_ratio(input_value: _T, expected: _T):
    assert whole_to_ratio(input_value) == expected



import copy

from django.contrib import admin
from django.db import models
from django.db.models.fields.proxy import OrderWrt
from model_utils.fields import StatusField
from simple_history.manager import HistoryManager
from simple_history.models import HistoricalRecords as BaseHistoricalRecords
from simple_history.models import transform_field

from . import override


class HistoricalRecords(BaseHistoricalRecords):
    """Custom HistoricalRecords to handle StatusField and OrderWrt."""

    @override
    def copy_fields(self, model):
        """Add handling for StatusField."""
        fields = {}
        for og_field in self.fields_included(model):
            field = copy.copy(og_field)
            field.remote_field = copy.copy(field.remote_field)
            if isinstance(field, StatusField):
                field.__class__ = models.CharField  # type: ignore[assignment]
            if isinstance(field, OrderWrt):
                # OrderWrt is a proxy field, switch to a plain IntegerField
                field.__class__ = models.IntegerField  # type: ignore[assignment]
            if isinstance(field, models.ForeignKey):
                old_field = field
                old_swappable = old_field.swappable
                old_field.swappable = False
                try:
                    _name, _path, args, field_args = old_field.deconstruct()
                finally:
                    old_field.swappable = old_swappable
                if getattr(old_field, "one_to_one", False) or isinstance(
                    old_field, models.OneToOneField
                ):
                    FieldType = models.ForeignKey  # noqa: N806
                else:
                    FieldType = type(old_field)  # noqa: N806

                # Remove any excluded kwargs for the field.
                # This is useful when a custom OneToOneField is being used that
                # has a different set of arguments than ForeignKey
                for exclude_arg in self.field_excluded_kwargs(old_field):
                    field_args.pop(exclude_arg, None)

                # If field_args['to'] is 'self' then we have a case where the object
                # has a foreign key to itself. If we pass the historical record's
                # field to = 'self', the foreign key will point to an historical
                # record rather than the base record. We can use old_field.model here.
                if field_args.get("to", None) == "self":
                    field_args["to"] = old_field.model

                # Override certain arguments passed when creating the field
                # so that they work for the historical field.
                field_args.update(
                    db_constraint=False,
                    related_name="+",
                    null=True,
                    blank=True,
                    primary_key=False,
                    db_index=True,
                    serialize=True,
                    unique=False,
                    on_delete=models.DO_NOTHING,
                )
                field = FieldType(*args, **field_args)
                field.name = old_field.name
            else:
                transform_field(field)

            # drop db index
            if field.name in self.no_db_index:
                field.db_index = False

            fields[field.name] = field
        return fields


class HistoryStampManager(HistoryManager):
    def created(self):
        return self.order_by("history_date").first()

    def modified(self):
        return self.order_by("-history_date").first()


class HistoryModel(models.Model):
    """Adds useful tracking fields.

    Properties:
      created_by (str): history_user from first history record
      created (datetime): history_date from first history record
      modified_by (str): history_user from last history record
      modified (datetime): history_date from last history record
    """

    history = HistoricalRecords(history_manager=HistoryStampManager, inherit=True)

    class Meta:
        abstract = True

    @property
    @admin.display
    def created_by(self):
        return self.history.created().history_user

    @property
    @admin.display
    def created(self):
        return self.history.created().history_date

    @property
    @admin.display
    def modified_by(self):
        return self.history.modified().history_user

    @property
    @admin.display
    def modified(self):
        return self.history.modified().history_date



"""Check that basic features work.

Catch cases where e.g., files are missing so the import doesn't work. It is recommended
to check that e.g., assets are included."""

from django_rubble.utils.numeric_utils import is_number


def test_basic_imports():
    """Test that basic imports and functions work."""
    return is_number(10)


if __name__ == "__main__":
    try:
        from rich import print

        msg = "[bold green]🎉 Smoke test succeeded 🎉[/bold green]"
    except ImportError:
        msg = "Smoke test succeeded"

    if test_basic_imports():
        print(msg)
    else:
        raise RuntimeError("Smoke test failed")



"""Django Rubble app configuration."""

from django.apps import AppConfig


class UtilsConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_rubble"
    verbose_name = "Django Rubble"



from pydantic import BaseModel, RootModel


class PydanticBase(BaseModel):
    pass


class PydanticRootBase(RootModel):
    root: list[PydanticBase]

    def __getitem__(self, item):
        return self.root[item]

    def __len__(self):
        return len(self.root)

    def __iter__(self):
        return iter(self.root)



from datetime import UTC, datetime

from croniter import croniter
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from django_rubble.forms.db_forms import (
    SimplePercentageField as SimplePercentageFormField,
)


class SimplePercentageField(models.DecimalField):
    """Enter and display percentages out of 100 but store them out of 1
    in db as decimals

    Because this is based on `models.DecimalField`, `decimal_places` applies to what is
    stored in the db (/1), not what is shown or typed in (/100). With that said, add two
    (2) to whatever is desired in the form for proper validation.
    """

    description = _(
        "percentage (max {max_digits} digits; {decimal_places} decimal places"
    )
    log_name = "models.PercentageField"

    def __init__(
        self,
        verbose_name=None,
        name=None,
        max_digits=None,
        decimal_places=None,
        **kwargs,
    ):
        if decimal_places is not None:
            self.humanize_decimal_places = int(decimal_places) - 2
        else:
            self.humanize_decimal_places = None

        kwargs.update(
            {
                "max_digits": max_digits,
                "decimal_places": decimal_places,
            }
        )

        super().__init__(verbose_name, name, **kwargs)

    def formfield(self, **kwargs):
        defaults = {"form_class": SimplePercentageFormField}
        if self.decimal_places is not None:
            kwargs.update(decimal_places=self.decimal_places - 2)

        defaults.update(kwargs)

        return super().formfield(**defaults)


CRON_REGEX = r"^(\*|([0-5]?[0-9])) (\*|([0-5]?[0-9])) (\*|([01]?[0-9]|2[0-3])) (\*|([01]?[0-9]|2[0-3])) (\*|([0-3]?[0-9])) (\*|([01]?[0-9]|2[0-3])) (\*|([0-5]?[0-9])) (\*|([01]?[0-9]|2[0-3]))$"  # noqa: E501


def validate_cron_expression(value):
    try:
        _ = croniter(value, datetime.now(tz=UTC))
    except (ValueError, KeyError):
        msg = f"{value} is not a valid cron expression."
        raise ValidationError(msg) from None


class CronExpressionField(models.CharField):
    default_validators = [validate_cron_expression]
    description = "A field to store a cron expression"

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = kwargs.get(
            "max_length", 100
        )  # Set max_length to 100 if not provided
        super().__init__(*args, **kwargs)



from contextlib import nullcontext

import pytest
from django.core.exceptions import ValidationError

from django_rubble.fields.db_fields import CronExpressionField, validate_cron_expression


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("0 1 * * 5", nullcontext(enter_result=None)),
        ("* * 1", pytest.raises(ValidationError)),
    ],
)
def test_cron_validation(value, expected):
    with expected as e:
        assert validate_cron_expression(value) == e


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("0 1 * * 5", nullcontext(enter_result="0 1 * * 5")),
        ("* * 1", pytest.raises(ValidationError)),
    ],
)
def test_cron_field(value, expected):
    field = CronExpressionField()
    with expected as e:
        assert field.clean(value, None) == e



import pytest

from django_rubble.utils.strings import truncate_string


def test_truncate_string_basic():
    assert truncate_string("This is a long string", 10) == "This is..."
    assert truncate_string("This is a long string", 10, postfix="..") == "This is.."
    assert (
        truncate_string("This is a long string", 20, postfix="..")
        == "This is a long str.."
    )
    assert truncate_string("This is a long string", 25) == "This is a long string"


def test_truncate_string_no_truncation():
    assert truncate_string("Short", 10) == "Short"
    assert truncate_string("Exact length", 12) == "Exact length"


def test_truncate_string_with_spaces():
    assert truncate_string("This is a long string ", 10) == "This is..."
    assert truncate_string("This is a long string ", 10, postfix="..") == "This is.."
    assert (
        truncate_string("This is a long string ", 20, postfix="..")
        == "This is a long str.."
    )


@pytest.mark.parametrize(
    ("string", "num_char", "expected"),
    [
        ("", 10, ""),
        (" ", 10, ""),
        ("A", 1, "A"),
        ("A", 2, "A"),
        ("A ", 2, "A"),
    ],
)
def test_truncate_string_edge_cases(string, num_char, expected):
    assert truncate_string(string, num_char) == expected


def test_truncate_string_postfix_longer_than_num_char():
    assert truncate_string("This is a long string", 5, postfix="......") == "....."


if __name__ == "__main__":
    pytest.main()



import requests
from attrmagic import SimpleRoot
from furl import furl


class RuleSelector(SimpleRoot[str]): ...


json_to_python_type: dict[str, type] = {
    "string": str,
    "number": float,
    "integer": int,
    "boolean": bool,
    "array": list,
    "object": dict,
}


def generate_ruff_schema(url: furl | str):
    if not isinstance(url, furl):
        url = furl(url)
    schema_json = requests.get(url, timeout=10).json()

    return RuleSelector.model_validate(
        schema_json["definitions"]["RuleSelector"]["enum"]
    )



"""Message functions for the checks module.

This module provides functions that return error messages for the checks module.

Available functions:

- must_be: Return a list of errors for a missing or incorrect option value.
- refer_to_missing_field: Return a list of errors for a missing field."""

from django.core import checks
from django.db import models

from django_rubble.utils.model_helpers import get_model_label


def must_be(
    type_description: str, option: str, obj: object, error_id: str
) -> list[checks.Error]:
    """Return a list of errors for a missing or incorrect option value.

    Examples:
        >>> must_be("a list", "natural_key_fields", cls, "rubble.E002")
        [Error: The value of 'natural_key_fields' must be a list.]

    Args:
        type_description (str): The description of the required type.
        option (str): The name of the option.
        obj (object): The object to which the option belongs.
        error_id (str): The error ID.

    Returns:
        list[Error]: A list of errors
    """
    return [
        checks.Error(
            f"The value of '{option}' must be {type_description}.",
            obj=obj.__class__,
            id=error_id,
        ),
    ]


def refer_to_missing_field(
    field_name: str, option: str, obj: type[models.Model], error_id: str
) -> list[checks.Error]:
    """Return a list of errors for a missing field.

    Examples:
        >>> refer_to_missing_field("foo", "bar", MyModel, "rubble.E003")
        [Error: The value of 'bar' refers to 'foo', which is not a field of 'MyModel'.]

    Args:
        field_name (str): The name of the missing field.
        option (str): The name of the option.
        obj (type[models.Model]): The model to which the option belongs.
        error_id (str): The error ID.

    Returns:
        list[Error]: A list of errors
    """
    return [
        checks.Error(
            f"The value of '{option}' refers to '{field_name}',"
            f" which is not a field of '{get_model_label(obj)}'.",
            obj=obj,
            id=error_id,
        ),
    ]



import ast
from pathlib import Path
from typing import override

from loguru import logger
from rich import print as rich_print


class SecretsFinder(ast.NodeVisitor):
    def __init__(self):
        self.secrets = []

    @override
    def visit_Call(self, node):
        if isinstance(node.func, ast.Attribute):
            if node.func.attr == "get_secret":
                for keyword in node.args:
                    value = keyword.value
                    logger.info(f"Found secret: {value}")
                    self.secrets.append(value)
        self.generic_visit(node)


def find_secrets_in_settings(settings_file: str):
    settings_file_path = Path(settings_file)
    with settings_file_path.open("r") as file:
        tree = ast.parse(file.read(), filename=settings_file)

    finder = SecretsFinder()
    finder.visit(tree)
    return finder.secrets


def create_env_sample(secrets: list[str], output_file: str):
    output_file_path = Path(output_file)

    with output_file_path.open("w") as file:
        for secret in secrets:
            file.write(f"{secret.upper()}=\n")


if __name__ == "__main__":
    settings_file_path = "C:/Users/KW131407/repos/homebin/config/settings/production.py"
    output_file_path = ".env.sample"

    secrets = find_secrets_in_settings(settings_file_path)
    rich_print(f"Secrets found in settings file: {secrets}")
    create_env_sample(secrets, output_file_path)
    rich_print(f"Sample .env file created at {output_file_path}")



# ruff: noqa: T201

from pathlib import Path

BASE_DIR = Path(__file__).resolve(strict=True).parent.parent
APPS_DIR = BASE_DIR

print(f"BASE_DIR: {BASE_DIR}")

DEBUG = True

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.admin",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_rubble",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "db.sqlite3",
    }
}

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]



from pathlib import Path
from re import Pattern
from typing import TYPE_CHECKING, Annotated, Any, Literal

import toml
from caseconverter import kebabcase
from Levenshtein import distance
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ImportString,
    PlainSerializer,
    PlainValidator,
    RootModel,
)
from pydantic_settings import (
    BaseSettings,
    PyprojectTomlConfigSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)
from rich.theme import Theme

from scripts.ruff_rule_models import generate_ruff_schema

if TYPE_CHECKING:
    from pydantic_settings import PydanticBaseSettingsSource

script_theme = Theme(
    {
        "linkish": "bright_magenta",
    }
)

levenshtein_threshold = 1

# Define file paths
pyproject_path = Path("pyproject.toml")
ruff_path = Path("ruff.toml")

ruff_schema_url = "https://json.schemastore.org/ruff.json"


PathStr = Annotated[
    Path,
    PlainSerializer(lambda x: str(x.as_posix()), return_type=str, when_used="json"),
]

rule_selector = generate_ruff_schema(ruff_schema_url)


def rule_selector_validation(value: Any) -> str:
    if not value.isupper():
        msg = "RuleSelector must be all uppercase"
        raise ValueError(msg)
    if value not in rule_selector:
        closest = min(rule_selector, key=lambda x: distance(value, x))
        msg = f'"{value}" is not a valid RuleSelector, did you mean "{closest}"'
        raise ValueError(msg)
    return value


RuleSelector = Annotated[str, PlainValidator(rule_selector_validation)]


RuleList = Annotated[
    list[RuleSelector], PlainSerializer(lambda x: sorted(x), return_type=list)
]


class BaseConfig(BaseModel):
    model_config = ConfigDict(alias_generator=kebabcase, extra="forbid")


class AnalyzeConfig(BaseConfig):
    detect_string_imports: bool | None = None
    direction: Literal["dependents", "dependencies"] | None = None


class FormatConfig(BaseConfig):
    docstring_code_format: bool | None = None
    docstring_code_line_length: int | Literal["dynamic"] | None = None
    exclude: list[PathStr] | None = None
    indent_style: Literal["space", "tab"] | None = None
    line_ending: Literal["auto", "lf", "cr-lf", "native"] | None = None
    preview: bool | None = None
    quote_style: Literal["single", "double"] | None = None
    skip_magic_trailing_comma: bool | None = None


class LintExtendPerFileIgnoresConfig(RootModel[dict[PathStr, list[str]]]):
    pass


class LintPerFileIgnoresConfig(RootModel[dict[PathStr, list[str]]]):
    pass


class LintIsortConfig(BaseConfig):
    force_single_line: bool | None = None


class LintFlake8PytestStyleConfig(BaseConfig):
    fixture_parentheses: bool | None = None


class LintConfig(BaseConfig):
    allowed_confusables: list[str] | None = None
    dummy_variable_rgx: Pattern | None = None
    exclude: list[PathStr] | None = None
    explicit_preview_rules: bool | None = None
    extend_fixable: RuleList | None = None
    extend_ignore: (
        Annotated[RuleList, Field(deprecated="Interchangeable with `ignore`")] | None
    ) = None
    extend_per_file_ignores: LintExtendPerFileIgnoresConfig | None = None
    extend_safe_fixes: RuleList | None = None
    extend_select: RuleList | None = None
    extend_unsafe_fixes: RuleList | None = None
    external: RuleList | None = None
    fixable: Literal["ALL"] | RuleList | None = None
    ignore: RuleList | None = None
    ignore_init_module_imports: bool | None = None
    logger_objects: list[ImportString] | None = None
    per_file_ignores: LintExtendPerFileIgnoresConfig | None = None
    preview: bool | None = None
    select: RuleList | None = None
    task_tags: list[str] | None = None
    typing_modules: list[ImportString] | None = None
    unfixable: RuleList | None = None
    flake8_pytest_style: LintFlake8PytestStyleConfig | None = None
    isort: LintIsortConfig | None = None


class RuffConfig(BaseSettings):
    model_config = SettingsConfigDict(
        toml_file=ruff_path,
        pyproject_toml_table_header=("tool", "ruff"),
        alias_generator=kebabcase,
    )

    builtins: list[PathStr] | None = None
    cache_dir: PathStr | None = None
    exclude: list[PathStr] | None = None
    extend: PathStr | None = None
    extend_exclude: list[PathStr] | None = None
    extend_include: list[PathStr] | None = None
    fix: bool | None = None
    fix_only: bool | None = None
    force_exclude: bool | None = None
    include: list[PathStr] | None = None
    indent_width: int | None = None
    line_length: int | None = None
    namespace_packages: list[PathStr] | None = None
    output_format: str | None = None
    preview: bool | None = None
    required_version: str | None = None
    respect_gitignore: bool | None = None
    show_fixes: bool | None = None
    src: list[PathStr] | None = None
    target_version: str | None = None
    unsafe_fixes: bool | None = None
    analyze: AnalyzeConfig | None = None
    # TODO: Add more fields as needed
    format: FormatConfig | None = None
    lint: LintConfig | None = None

    def model_dump_toml(
        self, *, by_alias: bool = True, exclude_none: bool = True, **kwargs
    ):
        as_dict = self.model_dump(
            mode="json", by_alias=by_alias, exclude_none=exclude_none, **kwargs
        )
        return toml.dumps(as_dict)

    def write_file(self, file_path: Path = ruff_path):
        with file_path.open("w") as file:
            toml.dump(
                self.model_dump(mode="json", by_alias=True, exclude_none=True), file
            )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: "PydanticBaseSettingsSource",
        env_settings: "PydanticBaseSettingsSource",
        dotenv_settings: "PydanticBaseSettingsSource",
        file_secret_settings: "PydanticBaseSettingsSource",
    ) -> tuple["PydanticBaseSettingsSource", ...]:
        return (
            TomlConfigSettingsSource(settings_cls),
            PyprojectTomlConfigSettingsSource(settings_cls),
        )


if __name__ == "__main__":
    from rich.console import Console

    console = Console(theme=script_theme)

    console.print(ruff_path.resolve())

    assert ruff_path.exists()

    ruff_config = RuffConfig()

    console.print_json(ruff_config.model_dump_json(by_alias=True, exclude_none=True))



"""This is a cli tool."""

from pathlib import Path

import typer

from django_rubble.secrets.utilities import create_env_sample, find_secrets_in_settings

app = typer.Typer()


@app.command()
def create_sample(
    settings_file: Path,
    output_file: Path,
):
    """Create a sample .env file from a Django settings file."""

    secrets = find_secrets_in_settings(settings_file)
    create_env_sample(secrets, output_file)
    typer.echo(f"Sample .env file created at {output_file}")



"""Generate the code reference pages."""

from collections.abc import Iterable
from pathlib import Path

import mkdocs_gen_files
from loguru import logger

nav = mkdocs_gen_files.Nav()

root = Path(__file__).parent.parent
src = root

msg = f"src path: {src}"
logger.debug(msg)


def any_in_iterable(iterable: Iterable[str], values: Iterable[str]) -> bool:
    """Check if any of the values are in the iterable."""
    return any(value in iterable for value in values)


for path in sorted(src.rglob("*.py")):
    module_path = path.relative_to(src).with_suffix("")
    doc_path = path.relative_to(src).with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    if parts[-1] == "__init__":
        parts = parts[:-1]
        if len(parts) == 0:
            continue
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif "test" in parts or parts[-1] == "__main__":
        msg = f"Skipping {module_path}"
        logger.debug(msg)

        continue

    nav[parts] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as nav_file:
        identifier = ".".join(parts)
        print("::: " + identifier, file=nav_file)

    mkdocs_gen_files.set_edit_path(full_doc_path, path.relative_to(root))

literate_nav = nav.build_literate_nav()

msg = f"literate_nav: {literate_nav}"
logger.debug(msg)


with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(literate_nav)



from decimal import Decimal, DecimalTuple
from typing import TypeVar, overload, override

from pydantic import BaseModel


def is_number(s: object) -> bool:
    """Check if a value can be coerced into a number type.

    Examples:
        >>> is_number(10)
        True
        >>> is_number("hello")
        False
        >>> is_number(Decimal("3.14"))
        True

    Args:
        s (Any): The value to check.

    Returns:
        bool: True if the value can be coerced into a number type, False otherwise."""
    if s is None:
        return False
    try:
        _ = float(s)  # pyright: ignore[reportArgumentType]
    except ValueError:
        return False
    else:
        return True


def any_to_float(s: object, default: float = 0) -> float:
    """Cast value as float, return default if invalid type.

    Examples:
        >>> any_to_float("5.5", 2.3)
        5.5
        >>> any_to_float("test", 0)
        0

    Args:
        s (Any): The value to cast.

    Returns:
        float: The value as a float, or the default if the value is not a number."""
    if not is_number(default):
        msg = f"Default must be of type `float` [{default}]"
        raise TypeError(msg)
    try:
        value_float = float(s)  # pyright: ignore[reportArgumentType]
    except ValueError:
        value_float = default

    return value_float


T = TypeVar("T", int, float, Decimal)


@overload
def ratio_to_whole(ratio: int | float) -> float: ...
@overload
def ratio_to_whole(ratio: Decimal) -> Decimal: ...
def ratio_to_whole(ratio: T) -> T:
    """Convert a ratio to a whole number.

    This is useful for converting a ratio to a percentage.

    Examples:
        >>> ratio_to_whole(0.03)
        3
        >>> ratio_to_whole(Decimal("1"))
        100

    Args:
        ratio (Decimal, float, str): The ratio to be converted.

    Returns:
        The whole number.
    """
    if isinstance(ratio, Decimal):
        return ratio * Decimal("100")

    return ratio * 100


@overload
def whole_to_ratio(whole: int | float) -> float: ...
@overload
def whole_to_ratio(whole: Decimal) -> Decimal: ...
def whole_to_ratio(whole: int | float | Decimal) -> float | Decimal:
    """Convert a whole number to a ratio.

    This is useful for converting a percentage to a ratio.

    Examples:
        >>> whole_to_ratio(3)
        0.03
        >>> whole_to_ratio(100)
        1

    Args:
        whole (Decimal, float, str): The whole number to be converted.

    Returns:
        The ratio. Decimal
    """
    match whole:
        case Decimal():
            return whole / Decimal("100")
        case _:
            return whole / 100


def trim_trailing_zeros(value: float | Decimal | str) -> Decimal:
    """Remove trailing zeros from a decimal value.

    This is useful for ensuring that a value can be safely compared with another value.

    Examples:
        >>> trim_trailing_zeros(3.1400)
        Decimal('3.14')
        >>> trim_trailing_zeros(Decimal("3.1400"))
        Decimal('3.14')

    Args:
        value (float, Decimal, str): The value to be trimmed.

    Returns:
        The trimmed value. Decimal
    """
    return Decimal(str(value)).normalize()


def set_zero(value: T) -> Decimal:
    """Set a value to a true Decimal zero if it is zero.

    Examples:
        >>> set_zero(0)
        Decimal('0')
        >>> set_zero(0.0)
        Decimal('0')

    Args:
        value (int, float, str): The value to be checked.

    Returns:
        The value as a Decimal if it is zero, otherwise the original value."""
    decimal_from_string = Decimal(str(value))

    if decimal_from_string == Decimal(0):
        return Decimal()

    return decimal_from_string


# TODO: this should *not* be a pydantic model
class Percent(BaseModel):
    """A model for handling percentages.

    This model is designed to handle percentages in a way that is more accurate than
    using floats.

    Attributes:
        value (Decimal, float, str): The value of the percentage.
        per_hundred (Decimal, float, str): The value of the percentage out of 100.
        decimal_places (int): The number of decimal places to use.
        has_decimal_places (bool): Whether the value has decimal places.
    """

    value: Decimal | float
    per_hundred: Decimal | float | None = None
    decimal_places: int | None = None
    has_decimal_places: bool | None = None

    @override
    def model_post_init(self, __context: object) -> None:  # noqa: PYI063
        value_decimal = Decimal(str(self.value))
        new_value = trim_trailing_zeros(value_decimal)

        per_hundred_dec = trim_trailing_zeros(ratio_to_whole(value_decimal))

        if self.decimal_places is not None:
            new_value = round(new_value, self.decimal_places + 2)
            per_hundred_dec = round(per_hundred_dec, self.decimal_places)
            self.has_decimal_places = True
        else:
            self.has_decimal_places = False

        self.value = set_zero(new_value)
        self.per_hundred = set_zero(per_hundred_dec)

        super().model_post_init(__context)

    @classmethod
    def fromform(
        cls, val: Decimal, field_decimal_places: int | None = None
    ) -> "Percent":
        """Create Percent from human-entry (out of 100)

        Examples:
            >>> Percent.fromform(Decimal(3))
            Percent(value=Decimal('0.03'), per_hundred=3)
            >>> Percent.fromform(Decimal("100"))
            Percent(value=Decimal('1'), per_hundred=100)

        Args:
            val (Decimal): The value of the percentage.
            field_decimal_places (int): The number of decimal places to use.

        Returns:
            Percent: The percentage model.
        """
        ratio_decimal = whole_to_ratio(val)
        return cls(value=ratio_decimal, decimal_places=field_decimal_places)

    def __mul__(self, other: float | int | Decimal | object):
        """Multiply using the ratio (out of 1) instead of human-readable out of 100

        Examples:
            >>> Percent(0.03) * 100
            Decimal('3')
            >>> Percent(1) * 100
            Decimal('100')"""
        if not is_number(other):
            return NotImplemented
        assert isinstance(other, (float, int, Decimal))
        match self.value:
            case float() | int():
                return self.value * float(other)
            case Decimal():
                return self.value * Decimal(str(other))

    def __float__(self):
        return float(self.value)

    def as_tuple(self) -> DecimalTuple:
        """Return the value as a decimal tuple.

        Returns:
            DecimalTuple: The value as a decimal tuple."""
        if not isinstance(self.value, Decimal):
            return Decimal(str(self.value)).as_tuple()
        return self.value.as_tuple()

    def is_finite(self):
        match self.value:
            case float():
                raise NotImplementedError("is_finite not implemented for float values")
            case Decimal():
                return self.value.is_finite()
            case int():
                return True

    @override
    def __repr__(self) -> str:
        return f"Percentage('{self.value}', '{self.per_hundred}%')"

    @override
    def __str__(self):
        return f"{self.per_hundred}%"
